.tilix-search-entry {
    padding-bottom: 6px;
    padding-top: 6px;
}

.tilix-search-frame {
    border-bottom: 1px solid @borders;
}

.tilix-notification-counter {
    background-color: @theme_fg_color;
    color: @theme_bg_color;
    border: 1px solid @borders;
    padding: 1px 4px 1px 4px;
    font-size: 0.8em;
}

.tilix-tab-notification-counter {
    background-color: @theme_selected_bg_color;
    color: @theme_selected_fg_color;
    border: 1px solid @borders;
    /*
    padding: 2px 4px 2px 4px;
    margin-top: 2px;
    margin-bottom: 2px;
    */
    font-size: 0.8em;
}

.tilix-small-button {
    padding: 2px 4px;
}

.tilix-notebook-page {
    background-color: @theme_bg_color;
}

/* Used to give backgrounds to widgets that end up transparent
   due to transparent window */
.tilix-background {
    background-color: @theme_bg_color;
}

/* Only used when CSD is disabled */
.tilix-toolbar {
    background-color: @theme_bg_color;
    border-bottom: 1px solid @borders;
}

.tilix-notification-count {
    background-image: -gtk-gradient(radial, center center, 0, center center, 0.5, to(@theme_selected_bg_color), to(transparent)), -gtk-gradient(radial, center center, 0, center center, 0.5, to(rgba(255, 255, 255, 0.76923)), to(transparent));
    color: @theme_selected_fg_color;
    font-size: smaller;
}

.tilix-notification-count > label {
    color: @theme_selected_fg_color;
}

.tilix-session-index {
    background-image: -gtk-gradient(radial, center center, 0, center center, 0.5, to(@theme_fg_color), to(transparent)), -gtk-gradient(radial, center center, 0, center center, 0.5, to(rgba(255, 255, 255, 0.76923)), to(transparent));
    color: @theme_bg_color;
    font-size: smaller;
}

.tilix-session-index > label {
  color: @theme_bg_color;
}

.tilix-session-sidebar {
    background-color: @theme_bg_color;
}

revealer.left > scrolledwindow {
    border-width: 0 1px 0 0;
}

revealer.right > scrolledwindow {
    border-width: 0 0 0 1px;
}

.tilix-terminal-scrolledwindow > undershoot.top {
    background-image: none;
}

.tilix-terminal-scrolledwindow > undershoot.bottom {
    background-image: none;
}

.tilix-session-name {
    color: @theme_bg_color;
    background-color: @theme_fg_color;
    border-radius: 8px 8px 8px 8px / 50% 50% 50% 50%;
    padding-left: 8px;
    padding-right: 8px;
    font-size: smaller;
}

/* Remove padding from buttons when in compact mode */
.compact > .button {
	padding: 0px;
}

/* Support small titlebars in 3.20 */
.compact > button {
	padding: 0px;
}

.tilix-sidebar-close-button {
    border: 0px;
    padding: 0px;
    margin: 0px;
}

.tilix-session-switcher {
    background-color: @theme_bg_color;
}

/* Visual Bell Animation, note each frame defined
   separately because GTK doesn't support grouping them
   together like normal CSS
 */
@keyframes shake {
    0%   {-gtk-icon-transform: translateX(0);}
    100% {-gtk-icon-transform: translateX(0);}
    10%  {-gtk-icon-transform: translateX(-1px);}
    30%  {-gtk-icon-transform: translateX(-1px);}
    50%  {-gtk-icon-transform: translateX(-1px);}
    70%  {-gtk-icon-transform: translateX(-1px);}
    90%  {-gtk-icon-transform: translateX(-1px);}
    20%  {-gtk-icon-transform: translateX(1px);}
    40%  {-gtk-icon-transform: translateX(1px);}
    60%  {-gtk-icon-transform: translateX(1px);}
    80%  {-gtk-icon-transform: translateX(1px);}
}

.tilix-bell {
  background: none;
  opacity: 0;
  -gtk-icon-source: -gtk-icontheme("alarm-symbolic");
}

.tilix-bell:checked {
    opacity: 1;
    animation: shake 1s linear infinite;
}

/* New Button in new UI */
.session-new-button {
  padding-left: 2px;
  padding-right: 2px;
}

/* Remove rounded corners when disable-csd set to true */
.tilix-embedded-headerbar {
    border-radius: 0;
}

/* Quake window border */
.tilix-quake-frame {
    border: 1px solid rgba(0, 0, 0, .7);
}

/*
.tilix-quake-frame > border {
    border-width: 0px;
}
*/

.tilix-error {
    color: @error_color;
}