/* Easier navigation in tree view */
@binding-set TilixTreeViewBinding
{
    bind "Left" {
      "select-cursor-parent" ()
      "expand-collapse-cursor-row" (0,0,0)
    };
    bind "Right" {
      "expand-collapse-cursor-row" (0,1,0)
    };
}

treeview
{
    -gtk-key-bindings: TilixTreeViewBinding;
}