/*
 * This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0. If a copy of the MPL was not
 * distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */
module gx.tilix.terminal.actions;

enum ACTION_PREFIX = "terminal";
enum ACTION_FIND = "find";
enum ACTION_FIND_PREVIOUS = "find-previous";
enum ACTION_FIND_NEXT = "find-next";
enum ACTION_LAYOUT = "layout";
enum ACTION_CLOSE = "close";
enum ACTION_MAXIMIZE = "maximize";
enum ACTION_ENCODING_SELECT = "encoding-select";
enum ACTION_PROFILE_PREFERENCE = "profile-preference";
enum ACTION_ENCODING_PREFERENCE = "encoding-preference";
enum ACTION_READ_ONLY = "read-only";
enum ACTION_RESET = "reset";
enum ACTION_RESET_AND_CLEAR = "reset-and-clear";
enum ACTION_COPY = "copy";
enum ACTION_COPY_AS_HTML = "copy-as-html";
enum ACTION_PASTE = "paste";
enum ACTION_PASTE_PRIMARY = "paste-primary";
enum ACTION_ADVANCED_PASTE = "advanced-paste";
enum ACTION_COPY_LINK = "copy-link";
enum ACTION_OPEN_LINK = "open-link";
enum ACTION_SELECT_ALL = "select-all";
enum ACTION_UNSELECT_ALL = "unselect-all";
enum ACTION_ZOOM_IN = "zoom-in";
enum ACTION_ZOOM_OUT = "zoom-out";
enum ACTION_ZOOM_NORMAL = "zoom-normal";
enum ACTION_SAVE = "save";
enum ACTION_SYNC_INPUT_OVERRIDE = "sync-input-override";
enum ACTION_INSERT_NUMBER = "insert-number";
enum ACTION_INSERT_PASSWORD = "insert-password";
enum ACTION_TITLE_STYLE = "title-style";
enum ACTION_SELECT_BOOKMARK = "select-bookmark";
enum ACTION_ADD_BOOKMARK = "add-bookmark";
enum ACTION_SCROLL_UP = "scroll-up";
enum ACTION_SCROLL_DOWN = "scroll-down";
enum ACTION_PAGE_UP = "page-up";
enum ACTION_PAGE_DOWN = "page-down";
enum ACTION_MONITOR_SILENCE = "monitor-silence";
enum ACTION_FILE_BROWSER = "file-browser";
enum ACTION_PREVIOUS_PROMPT = "previous-prompt";
enum ACTION_NEXT_PROMPT = "next-prompt";
enum ACTION_TOGGLE_MARGIN = "toggle-margin";